
#include <qworkspace.h>
#include <qframe.h>
#include <qlayout.h>
#include <qwidget.h>
#include <qsize.h>
#include <qpoint.h>
#include <qcolor.h>
#include <qstatusbar.h>
#include <qevent.h>

#include "GenericWorkspace.h"

#include <s11n.net/s11n/s11n_debuggering_macros.hpp> // CERR
#include <iostream> // needed by CERR

namespace gws {

        GenericWorkspaceWindow::GenericWorkspaceWindow( QWidget * child, QWidget * parent )
                : QFrame( parent, "GenericWorkspaceWindow", Qt::WDestructiveClose ), m_child(child)
        {
                this->init();
        }
        GenericWorkspaceWindow::~GenericWorkspaceWindow()
        {
//                 CERR << "~GenericWorkspaceWindow()\n";
        }

        void GenericWorkspaceWindow::init()
        {
                QLayout * layout = new QHBoxLayout( this );
                layout->setAutoAdd( true );
                layout->setMargin( 0 );
                layout->setSpacing( 0 );
                if( QWidget * ch = this->m_child )
                {
                        ch->reparent( this, QPoint(0,0), true );
                        connect( ch, SIGNAL(destroyed()), this, SLOT(deleteLater()) );
                        int foo = 0;
                        int mx = 3000;
                        setMaximumSize( (foo=ch->maximumSize().width()+10)>mx
                                        ? mx
                                        : foo, (foo=ch->maximumSize().height()+10)>mx ? mx : foo );
                        setMinimumSize( ch->minimumSize() );
                }
        }

        void GenericWorkspaceWindow::closeEvent( QCloseEvent *e )
        {
                e->accept();
//                 hide();
//                 delete( this->m_child );
        }

        void GenericWorkspace::init()
        {
                this->setBackgroundColor( QColor( 0, 0, 0x7f ) );
                this->setScrollBarsEnabled( true );
        }

        QSize
        GenericWorkspaceWindow::sizeHint()
        {
                if( ! this->m_child ) return QFrame::sizeHint();
                return this->m_child->size();
        }

        void
        GenericWorkspaceWindow::showMaximized()
        {
                this->QFrame::showMaximized();
                if( ! this->m_child ) return;
                QSize s = size();
                QSize m = this->m_child->maximumSize();
                resize( s.width() > m.width() ? m.width() : s.width(),
                        s.height() > m.height() ? m.height() : s.height()
                        );
        }

        GenericWorkspace::GenericWorkspace( QWidget * parent )
                : QWorkspace( parent, "GenericWorkspace" )
        {

        }

        GenericWorkspace::~GenericWorkspace()
        {
//                 CERR << "qobjects lefts: " << this->qobjects().size() << "\n";

        }

        void GenericWorkspace::childDying( QObject * obj )
        {
//                 CERR << "childDying()" << "\n";
                qobject_map::iterator it = this->qobjects().find( obj );
                if( it == this->qobjects().end() ) return;
//                 CERR << "no, really: childDying()" << "\n";
                this->qobjects().erase( it );
        }

        GenericWorkspace::qobject_map &
        GenericWorkspace::qobjects()
        {
                return this->m_qobjects;
        }


        void GenericWorkspace::addWidget( QWidget * w )
        {
                GenericWorkspaceWindow * ch = new GenericWorkspaceWindow( w, this );
                ch->setCaption( w->className() );
                ch->reparent( this, ch->pos(), true );
                pair_type & pair = this->qobjects()[w];
                pair.first = w;
                pair.second = ch;
                this->connect( w, SIGNAL( destroyed(QObject *) ), SLOT( childDying( QObject * ) ) );
        }



} // namespace gws

