#ifdef NDEBUG
#  undef NDEBUG // force assert() to work
#endif
#include <cassert>
#include <memory> // auto_ptr

#include <stdlib.h> // getenv()

#include <s11n.net/s11n/s11nlite.hpp>
#include <s11n.net/s11n/list.hpp>
#include <s11n.net/s11n/pods_streamable.hpp>
#include <s11n.net/s11n/s11n_debuggering_macros.hpp>
#include <s11n.net/acme/argv_parser.hpp>
#include <s11n.net/cl/dll_util.hpp>

#include <qt.h>
#include <qsize.h>
#include <qpoint.h>

#include "s11nbrowser_config.hpp"
#include "S11nNodeListViewItem.h"
#include "s11nbrowserwindow.h" // Qt-generated file

#include <s11n.net/s11n/node_tree.hpp>

#include "qt_serialize.hpp"

typedef s11n::util::node_tree<s11nlite::node_type> TreeUtil;
TreeUtil & node_tree()
{
        static TreeUtil bob;
        return bob;
}

QListView * m_list = 0;
QListView & node_list()
{
        return *m_list;
}

s11nlite::node_type * m_root = 0;
s11nlite::node_type * root_node()
{
        return m_root;
}

/**
   Deletes previous root_node(), clears node_list() and node_tree(),
   and sets new root to n. Should return the total number of nodes
   under n, including n.
*/
size_t root_node( s11nlite::node_type * n )
{
        if( n != m_root)
        {
                delete m_root;
                m_root = n;
        }
        node_tree().clear();
        node_list().clear();
        size_t c = 0;
        if( n )
        {
                c = node_tree().climb_tree( n );
                typedef s11n::qt::S11nNodeListViewItem NLI;
                new NLI( *n, &node_list() ); // node_list() owns the new NLI.
                node_list().firstChild()->setOpen( true );
        }
        return c;
}


bool steal_node( s11nlite::node_type * node )
{
        if( ! node ) return false;
        return node_tree().take_node( node );
//        return 0 != node_tree().climb_tree( root_node() );
}

s11nlite::node_type * load_node_file( const std::string & filename,
                                      const std::string & serclass )
{

        typedef s11nlite::node_type NT;
        
        typedef s11nlite::serializer_base_type SER;
        typedef std::auto_ptr<SER> APS;

        NT * node = 0;
        if( serclass.empty() )
        {
                node = s11nlite::load_node( filename );
        }
        else
        {
                APS ser = APS( s11nlite::create_serializer( serclass ) );
                if( ! ser.get() )
                {
                        CERR << "load_node_file("<<filename<<","<<serclass<<"): couldn't classload requested Serializer type.\n";
                        return false;
                }
                node = ser->deserialize( filename );
        }
        if( ! node  )
        {
                CERR << "load_node_file("<<filename<<","<<serclass<<") failed to load a node :(\n";
                return 0;
        }
//         size_t count = 
//         node_tree().climb_tree( root_node() );
//         CERR << count << " node"<<(count==1 ? "" : "s")<<" read in.\n";

//          typedef s11n::qt::S11nNodeListViewItem NLI;
//         new NLI( * root_node(), &node_list() ); // node_list() owns the new NLI.
        return node;
}

struct root_node_cleaner
{
        root_node_cleaner() {}
        ~root_node_cleaner()
        {
                delete root_node();
        }

};


/**
   Loads Serializer DLLs.
*/
void autoload_serializers()
{
        typedef std::list<std::string> SList;
        SList li;
#define SER(X) li.push_back( "s11n/" # X "_serializer" );
        SER(mysql);
#undef SER

        SList::iterator it = li.begin(), et = li.end();
        for( ; et != it; ++it )
        {
                try
                {
                        cllite::open_dll( *it );
                }
                catch( const cllite::cl_exception & ex )
                {
                        CERR << "Autoload of '"<<(*it)<<"' failed. :(\n"
                             << ex.what() << "\n";
                        continue;
                }
        }
}

#define CFGKEY_SIZE "s11nbrowser_size"

int main( int argc, char ** argv )
{
        enum {
        ErrorDLL
        };

        typedef s11nlite::node_traits TR;
//         s11nlite::node_type cfg;

        root_node_cleaner cleaner;
        acme::argv_parser & args = acme::argv_parser::args( argc, argv );

        cllite::use_exceptions( true );

        /////////////////////////////////////////////////////////////////////
        // load requested DLLs
        typedef std::list<cllite::dll::dlentry> DList;
        DList dlllist;
        if( 0 != cllite::dll::parse_dlentry_flags( argc -1 , argv+1, dlllist ) )
        {
                try
                {
                        cllite::dll::dlload( dlllist );

                }
                catch( const cllite::cl_exception & ex )
                {
                        CERR << ex.what() << "\n";
                        return ErrorDLL;
                }
                if( 0 )
                {
                        CERR << "dlentries:\n";
                        s11nlite::save( dlllist, std::cout );
                        DList delist;
                        s11nlite::s11n_cast( dlllist, delist );
                        s11nlite::save( delist, std::cout );
                }
        }
        autoload_serializers();
        // end DLL loading.
        ////////////////////////////////////////////////////////////////////////


        args.set_help( "f FILENAME", "Loads the given file on startup." );
        args.set_help( "?, help", "Show this help." );
        args.set_help( "dl dllname",
                       std::string("Opens the given DLL. ")
                       +"May be used multiple times. All DLLs are loaded before the Serializer (-s) is loaded, "
                       +"DLL names may be absolute file names or base filenames if those files are in the lib path. "
                       +"This is useful when, e.g., the Serializer treats filenames as db keys or something similar."
                       +" Use -v to see information about what DLLs get loaded this way."
                       );

        args.set_help( "DL dllname",
                       std::string("Exacly like -dl")+
                       ", but does not abort if the dll is not found or if there is an error opening it. "+
                       "In s11n slang this is called a \"tolerant dll\"." );

        args.set_help( "s", "Sets the default s11nlite Serializer (i.e., the default output format)." );

        if( args.is_help_set() )
        {
                std::cout << s11n_BROWSER_NAME << " Version " << s11n_BROWSER_VERSION
                          << "\nLicense: " << s11n_BROWSER_LICENSE_TEXT
                          << "\nCommand-line options:\n\n";
                std::cout << args.dump_help();
                return 0;
        }

        std::string serclass = args.get( "s", "" );
        if( ! serclass.empty() )
        {
                s11nlite::serializer_class( serclass );
        }





        QApplication app( argc, argv );

// When using a db i get this bogus error from qt:
//     QSqlDatabase: QMYSQL3 driver not loaded
//     QSqlDatabase: available drivers: QMYSQL3 QPSQL7
// But the db access works fine via Qt Designer.
//
// #define DB_MYS11N_DRIVER     "QMYSQL3"
// #define DB_MYS11N_DBNAME     "mys11n"
// #define DB_MYS11N_USER       ""
// #define DB_MYS11N_PASSWD     ""
// #define DB_MYS11N_HOST       ""
//         QSqlDatabase *defaultDB = QSqlDatabase::addDatabase( DB_MYS11N_DRIVER );
//         defaultDB->setDatabaseName( DB_MYS11N_DBNAME );
//         defaultDB->setUserName( DB_MYS11N_USER );
//         defaultDB->setPassword( DB_MYS11N_PASSWD );
//         defaultDB->setHostName( DB_MYS11N_HOST );
//         if ( ! defaultDB->open() ) {
//                 CERR << "Db open failed :(\n";
//         }


        S11nBrowserWindow * w = new S11nBrowserWindow;
        w->setCaption( QString("s11n Data Node Browser ver. ") + s11n_BROWSER_VERSION );
        app.setMainWidget( w );
        m_list = w->listNodes;
        QSize wsz = s11nlite::config().get( CFGKEY_SIZE, QSize(600,500) );
        w->resize( wsz );
        w->show();

        std::string fname = args.get( "f", "" );
        if( !  fname.empty() )
        {
                size_t count = root_node( load_node_file( fname, "" ) );
                CERR << count << " node" << (count==1 ? "" : "s") << " loaded from "<<fname<<".\n";
        }


        int ret = app.exec();
        wsz = w->size();
        s11nlite::config().set( CFGKEY_SIZE, wsz  ); // i'm surprised that w is still valid here.
        root_node( 0 );
        delete( w );
        return ret;
}

